/*
 *	Acknowledgements:
 *		The first version of this DMC1800 driver was
 *		based on the DMC1842 driver by R. Treffers, UCB.
 *
 *	References:
 *	      Rubini, Linux Device Drivers O-Reilly
 *	      Galil DMC-1700/1800 User Manual Re.v. 1.2b Chapter 4
 *	      * Warning: the section on secondary FIFO handling is
 *	      * completely wrong for the DMC-1800!
 *
 */

#define MODULE
#define __KERNEL__

#include <asm/io.h>    
#include <asm/spinlock.h> 
#include <asm/uaccess.h> 
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>

#include "dmc1800.h"

#ifndef DMC_1800_MAJOR
#define DMC_1800_MAJOR 28			/* major device number */
#endif

#define ModuleName "dmc1800"

/* PCI Device and Vendor codes for the DMC1800 */
#define DEVICE_ID		0x9050
#define VENDOR_ID		0x10B5
#define SUBSYSTEM_DEVICE_ID	0x1800
#define SUBSYSTEM_VENDOR_ID	0x1079

/*
 * Describes a single DMC1800, with a link to another DMC1800 device
 * for using in linking such devices...
 */
#define NJIFFIEHIST 20	/* make sure that this isn't so large that the /proc
			 * output becomes too large.
			 */
typedef struct DMC1800 {
    struct wait_queue *a_inq, *b_inq;	/* ascii read, binary read queues */
    struct wait_queue *outq, *pollq;	/* write, poll queues */
    struct pci_dev *pci;		/* pci device info */
    struct tq_struct task;		/* for running the devicer poller */
    int task_alldone;			/* flags when task is not reenqueued */
    spinlock_t lock;			/* used in irq handler (at least) */
    struct dmc_stats stats;		/* for stats on this device */
    int instance;			/* which dmc1800 this is: 0,1,2,... */
    int n_open;				/* curr # open's on this instance;
    					 * by using this, we know what we are
					 * counting, whereas MOD_{INC,DEC}_USE
					 * depends on us being a module and
					 * doesn't guarantee what it counts.
					 */
    unsigned long last_bh_jiffies;	/* the bh iochecker sets this to the
    					 * current value of jiffies.
					 */
    unsigned long jiffiehist[NJIFFIEHIST];	/* Histogram of changes in
    					 * last_bh_jiffies.  (If > N*HIST,
					 * increment last bin).
					 */
    int base;				/* dmc1800 base address -- 
    					 * the address N in chapter 4
					 * of the Galil 1800 reference.
					 */
    unsigned char
    	interrupt[DMC1800_N_INTERRUPT+1];	/* The first element
    					 * is the number of entries in the
					 * queue; the succeeding values 
					 * are the interrupt values.
					 */
    struct DMC1800 *next;		/* next DMC1800 in list */
} DMC1800;

/*
 * Describes a single "open(2)" connection.  Multiple DMC1800_handles
 * can apply to a single DMC1800 instance -- one might be binary;
 * another ascii, etc; or someone might simply open the same device
 * multiple times by mistake.  Each of those gets one of these handles.
 */
typedef struct DMC1800_handle {
    DMC1800 *Dmc1800;
    int binary;				/* !0 == use binary interface */
    int need_halfbuffer;		/* !0 == dmc_poll() doesn't mark
    					 * device as writable unless queue
					 * is at least 1/2 empty.
					 */
} DMC1800_handle;

static DMC1800 *Dmc1800List = NULL;	/* All DMC1800 devices that we
					 * find at init time are listed
					 * here.
					 */


static int dmc_read(struct file *, char *, size_t, loff_t *);
static int dmc_write(struct file *, const char *, size_t, loff_t *);
static unsigned int dmc_poll(struct file *, poll_table *);
static int dmc_ioctl(struct inode *, struct file *,unsigned int,unsigned long);
static void dmc_bh_iochecker(void *);
static int dmc_read_proc(char *, char **, off_t, int, int *, void *);
static int dmc_read_ascii(struct file *, char *, size_t, loff_t *);
static int dmc_read_binary(struct file *, char *, size_t, loff_t *);
static int dmc_open(struct inode *, struct file *);
static int dmc_release(struct inode *, struct file *);
static void dmc_irq_handler(int irq, void *dev_id, struct pt_regs *regs);
static void dmc_clear_reg(DMC1800 *dev, unsigned long);
static void dmc_clear_stats(DMC1800 *dev);

struct file_operations dmc_fops = {
	NULL,           /* _seek */
	dmc_read, 	/* _read */
	dmc_write, 	/* _write */
	NULL,           /* _readdir */
	dmc_poll,       /* poll */
	dmc_ioctl,  	/* _ioctl */
	NULL,           /* _mmap */
	dmc_open,	/* _open */
	NULL,           /* _flush */
	dmc_release,   	/* _release */
	NULL,           /* _fsync */
	NULL,           /* _fasync */
	NULL,           /* _check_m */
	NULL,           /* _revalidate */
	NULL            /* _lock */
};

#ifdef USE_PROC_REGISTER
static int dmc_get_info(
    char *page, char **start, off_t offset, int count, int dummy)
{
    int dummy_eof;
    return dmc_read_proc(page, start, offset, count, &dummy_eof, NULL);
}

struct proc_dir_entry dmc_proc_status = {
        0,			/* low_ino: the inode -- dynamic */
        sizeof(ModuleName)-1, ModuleName,	/* len of name and name */
        S_IFREG | S_IRUGO,	/* mode */
        1, 0, 0,		/* nlinks, owner, group */
        0, NULL,		/* size - unused; operations -- use default */
        dmc_get_info,		/* function used to read data */
	NULL, NULL, NULL, NULL, NULL /* fill_inode, next, par, subdir, data */
    };

#endif

/* #undef DEBUG_OTHER */
/* #undef DEBUG_IOCTL */
/* #undef DEBUG_WRITE */
/* #undef DEBUG_READ */

int
init_module(void)
{
    short subvendor, subdevice;
    struct pci_dev *pdev;
    DMC1800 *dmcp;
    int n_instances, err;

#ifdef USE_PROC_REGISTER
    proc_register(&proc_root, &dmc_proc_status);
#else
    create_proc_read_entry(ModuleName, 0, NULL, dmc_proc_get_info, NULL);
#endif

    printk(KERN_INFO "DMC1800 driver compiled at %s %s\n",
    	__DATE__, __TIME__);

    if ((err = register_chrdev(DMC_1800_MAJOR, ModuleName, &dmc_fops)) < 0) {
	printk(KERN_INFO "DMC1800 ERROR register_chrdev() err=%d\n", err);
	return err;
    }
    printk(KERN_INFO "DMC1800 major=%d\n", DMC_1800_MAJOR);

    /*
     * Find all the DMC1800 cards
     */
    for (pdev=NULL, n_instances=0;
	    (pdev = pci_find_device(VENDOR_ID, DEVICE_ID, pdev)); ) {
	
	/* Check that subsystem vendor and device id's are DMC 1800 */
	if ((err=pci_read_config_word(pdev, 0x2c, &subvendor))) {
	    printk(KERN_INFO "DMC1800 driver err=%d getting "
				"subsystem vendor; skipping device\n", err);
	    continue;
	}
	if ((err=pci_read_config_word(pdev, 0x2e, &subdevice))) {
	    printk(KERN_INFO "DMC1800 driver err=%d getting "
				"subsystem device; skipping device\n", err);
	    continue;
	}
	if (subvendor != SUBSYSTEM_VENDOR_ID) {
	    printk(KERN_INFO "DMC1800 module: skipping device with "
				    "subsystem vendor=%#x\n", subvendor);
	    continue;
	}
	if (subdevice != SUBSYSTEM_DEVICE_ID) {
	    printk(KERN_INFO "DMC1800 module: skipping device with "
				    "subsystem device=%#x\n", subdevice);
	    continue;
	}

	if (!(dmcp = kmalloc(sizeof(DMC1800), GFP_KERNEL))) {
	    return -ENOMEM;
	}
	dmcp->instance = n_instances;	/* which DMC 1800 */
	dmcp->n_open = 0;		/* current num open's on this dev */
	dmcp->pci = pdev;

	dmcp->interrupt[0] = 0;		/* currently stored interrupt values */

	dmc_clear_stats(dmcp);		/* for keeping driver stats */

	printk(KERN_INFO "DMC1800 instance=%d irq=%d "
				"base=%lx %lx %lx %lx %lx %lx\n",
	    n_instances, pdev->irq,
	    pdev->base_address[0],
	    pdev->base_address[1],
	    pdev->base_address[2],
	    pdev->base_address[3],
	    pdev->base_address[4],
	    pdev->base_address[5]);

	/* Set up the FIFO address */
	dmcp->base = pdev->base_address[2] & PCI_BASE_ADDRESS_IO_MASK;

	/* Initialize the read/write wait queues */
	dmcp->a_inq = NULL;
	dmcp->b_inq = NULL;
	dmcp->outq = NULL;
	dmcp->pollq = NULL;

	request_region(dmcp->base, 16, ModuleName);

	spin_lock_init(&dmcp->lock);	/* useful in irq handler, etc */

	/* When the device is opened, we'll run this task on the immediate
	 * queue, so that we can poll for readable/writable.
	 */
	dmcp->task.sync = 0;
	dmcp->task.data = dmcp;
	dmcp->task.routine = dmc_bh_iochecker;
	dmcp->task_alldone = 0;		/* set !0 by bh_iochecker when all done;
					 * used at release time.
					 */

	/* Add this device to the list of dmc1800 devices */
	dmcp->next = Dmc1800List;
	Dmc1800List = dmcp;

	n_instances++;
    }
    if (n_instances == 0) {
	printk(KERN_INFO "DMC1800 driver can't find any PCI DMC1800 device\n");
	return -ENODEV;
    }
    return 0;
}

static void
dmc_clear_stats(DMC1800 *p) {
    p->stats.pollcount = 0;
    p->stats.aread = 0;
    p->stats.bread = 0;
    p->stats.areadsatisfied = 0;
    p->stats.breadsatisfied = 0;
    p->stats.aread_sleeping = 0;
    p->stats.bread_sleeping = 0;
    p->stats.readempty = 0;
    p->stats.rlastsleep = 0;
    p->stats.nwrite = 0;
    p->stats.nwritesatisfied = 0;
    p->stats.cantwrite = 0;
    p->stats.wlastsleep = 0;
    p->stats.write_sleeping = 0;
    p->stats.irq_caught = 0;
}

static void
dmc_irq_handler(
    int irq,
    void *dev_id,
    struct pt_regs *regs)
{
    DMC1800 *dev = (DMC1800 *) dev_id;
    DMC1800 *dmcp;
    unsigned char cntl_value, irq_value;
    int i;

    spin_lock(&dev->lock);

    /* Test if this is really one of ours (might not be if IRQ is
     * shared).  Do this be checking if the dev is in our devlist.
     */
    for (dmcp = Dmc1800List; dmcp && dmcp != dev; dmcp = dmcp->next)
	;
    if (!dmcp) {
	spin_unlock(&dev->lock);	/* not our IRQ */
	return;
    }

    /* Check if there's an IRQ pending */
    cntl_value = (unsigned char) inb(CNTL_ADDR(dev));
    if (!(cntl_value & DMC1800_IRQ_STATUS)) {
	/* hmm, no interrupt is pending.  Why not?! */
	spin_unlock(&dev->lock);
	return;
    }

    dev->stats.irq_caught++;	/* Number of IRQ's ever seen */

    /* In order to be able to receive another interrupt, we must
     * (a) read the IRQ register, and
     * (b) clear the IRQ bit in the control register.
     */
    irq_value = (unsigned char) inb(IRQ_ADDR(dev));
    if (dev->interrupt[0] < DMC1800_N_INTERRUPT) {
	/* Still have room to store interrupt values in our queue. */
	i = ++dev->interrupt[0];
	dev->interrupt[i] = irq_value;
    }
    outb_p(cntl_value | DMC1800_IRQ_CLEAR, CNTL_ADDR(dev));

    wake_up_interruptible(&dev->pollq);	/* procs mebbe waiting for interrupts */
    wake_up_interruptible(&dev->a_inq);	/* procs waiting on ascii read */
    wake_up_interruptible(&dev->b_inq);	/* procs waiting on binary read */
    wake_up_interruptible(&dev->outq);	/* procs waiting to try to write */

    spin_unlock(&dev->lock);
}

static void
dmc_bh_iochecker(
    void *data)
{
    DMC1800 *dev = data;
    int c, j;
    int woke_any = 0;

    dev->stats.pollcount++;

    /*
     * keep histogram of elapsed jiffies between calls to dmc_bh_iochecker
     */
    j = (jiffies - dev->last_bh_jiffies);
    if (j >= NJIFFIEHIST)
	j = NJIFFIEHIST - 1;
    dev->jiffiehist[j]++;
    dev->last_bh_jiffies = jiffies;

    /* primary FIFO readable? */
    if ((c=(inb(CNTL_ADDR(dev))) & DMC1800_READ_EMPTY) == 0) {
	wake_up_interruptible(&dev->a_inq);
	woke_any = 1;
    }

    /* secondary FIFO readable? secondary FIFO finished updating?
     * The first check is for processes that are waiting until there is
     * data in the secondary FIFO.  The second check is for processes
     * that have determined data is present and have set the freeze bit, and
     * are now waiting until the busy-bit is clear, so that they can safely
     * read the secondary FIFO.
     */
    if ((c & DMC1800_2ND_EMPTY) == 0 || (c & DMC1800_BUSY) == 0) {
	wake_up_interruptible(&dev->b_inq);
	woke_any = 1;
    }

    /* Writable? */
    if ((inb(CNTL_ADDR(dev)) & DMC1800_WRITE_FULL) == 0) {
	wake_up_interruptible(&dev->outq);
	woke_any = 1;
    }

    /* If any activity was worth waking up another queue, it's also
     * appropriate to wake up anybody on the poll queue.
     */
    if (woke_any) {
	wake_up_interruptible(&dev->pollq);
    }

    /* Re-queue ourselves as long as the device is open */
    if (dev->n_open > 0) {
	queue_task(&dev->task, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
    } else {
	/* Device is closed; indicate that we're all done with
	 * enqueueing ourselves.
	 */
	dev->task_alldone = 1;
    }

}

void cleanup_module(void)
{
    int err;
    struct DMC1800 *dev, *next;

#ifdef USE_PROC_REGISTER
    proc_unregister(&proc_root, dmc_proc_status.low_ino);
#else
    remove_proc_entry(ModuleName, 0);
#endif

    err=unregister_chrdev(DMC_1800_MAJOR, ModuleName);
    printk(KERN_INFO "DMC1800 unregister=%d\n",err);

    for (dev = Dmc1800List; dev; dev = next) {
	release_region(dev->base, 16);
	next = dev->next;
	kfree(dev);
    }
    Dmc1800List = NULL;
}

static
int dmc_open(
    struct inode *inode,
    struct file *file)
{
    DMC1800 *dev;
    DMC1800_handle *handle;

    int minor = MINOR(inode->i_rdev);
    int instance;
    int i;

    /* Bits 0..6 specify which DMC1800 */
    instance = (minor & 0x3F);

    /* Look for the device with that instance */
    for (dev = Dmc1800List;
    		dev && dev->instance != instance; dev = dev->next) {
	;
    }
    if (!dev) {
	/* No such card configured in */
	return -ENODEV;
    }

    if (!(handle = kmalloc(sizeof(DMC1800_handle), GFP_KERNEL))) {
	return -ENOMEM;
    }
    handle->Dmc1800 = dev;
    handle->binary = (minor & 128);	/* Bit 7 means use binary interface */
    handle->need_halfbuffer = 1;	/* 0 -> poll() says writable if
    					 * at least one byte can be written;
					 * !0 -> need at least 1/2 buffer
					 * (128 bytes) before poll() says
					 * it's writable.
					 */

    file->private_data = handle;

    dev->n_open++;
    if (dev->n_open == 1) {
	/* This is the first open, so enable IRQ's, add
	 * a periodic task to check for readable/writable,
	 * and create a /proc entry.
	 */
	for (i=0; i<NJIFFIEHIST; i++) {
	    dev->jiffiehist[i] = 0;
	}
	if (request_irq(dev->pci->irq, dmc_irq_handler, 0, ModuleName, dev)) {
	    printk(KERN_INFO "DMC1800 can't get assigned irq %d\n",
	    							dev->pci->irq);
	    dev->pci->irq = -1;

	} else {
	    /* Clear I/O and enable IRQ's */
	    dmc_clear_reg(dev,
	    		(DMC_CLEAR_READ | DMC_CLEAR_WRITE | DMC_CLEAR_IRQ));
	}

	dev->task_alldone = 0;			/* won't be cleared until
						 * task stops putting self on
						 * immediate queue.
						 */
	queue_task(&dev->task, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
    }

    MOD_INC_USE_COUNT;
    return 0;
}

static
int dmc_release(
    struct inode *inode,
    struct file *file)
{
    DMC1800_handle *handle = file->private_data;
    DMC1800 *dev;
    unsigned char c;

    dev = handle->Dmc1800;
    dev->n_open--;

    if (dev->n_open == 0) {
	/*
	 * Last open instance; disable IRQ's.
	 */
	if (dev->pci->irq >= 0) {
	    free_irq(dev->pci->irq, dev);
	    c = inb(CNTL_ADDR(dev)) | ~(DMC1800_IRQ_ENABLE);
	    outb_p(c, CNTL_ADDR(dev));
	}

	/*
	 * Wait for the task_alldone flag to become !0.  This will
	 * signal us that our immediate-queue task has stopped 
	 * re-enqueueing itself, and we can finish cleanup.
	 */
	do {
	    current->state = TASK_INTERRUPTIBLE;
	    schedule_timeout(1);
	} while (dev->task_alldone == 0);
    }

    kfree(handle);

    MOD_DEC_USE_COUNT;
    return 0;
}

static 
int dmc_ioctl(
    struct inode *inode, 
    struct file *file,
    unsigned int cmd, 
    unsigned long arg)
{
    int status, addr, n;
    struct dmc_iobuf iobuf;
    DMC1800_handle *handle = file->private_data;
    DMC1800 *dev = handle->Dmc1800;

    switch (cmd) {
    case DMC_CLEAR:
	dmc_clear_reg(dev, arg);
	break;
    case DMC_GET_CNTL:
	status=inb(CNTL_ADDR(dev));
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 get control/status reg=%x\n", status);
	#endif
	if (copy_to_user((void *)arg, (void *)&status, sizeof(int)))
	    return -EFAULT;
	break;
    case DMC_GET_FIFO:
	status=inb(FIFO_ADDR(dev));
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 get primary FIFO=%x\n", status);
	#endif
	if (copy_to_user((void *)arg, (void *)&status, sizeof(int)))
	    return -EFAULT;
	break;
    case DMC_GET_FIFO2:
	status=inb(SCND_FIFO_ADDR(dev));
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 get secondary FIFO=%x\n", status);
	#endif
	if (copy_to_user((void *)arg, (void *)&status, sizeof(int)))
	    return -EFAULT;
	break;
    case DMC_GET_IRQ:
	status=inb(IRQ_ADDR(dev));
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 get irq reg=%x \n", status);
	#endif
	if (copy_to_user((void *)arg, (void *)&status, sizeof(int)))
	    return -EFAULT;
	break;
    case DMC_RESET_CONTROLLER:
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 reset controller\n");
	#endif
	outb_p(DMC1800_RESET_CNTLR, IRQ_ADDR(dev));
	break;
    case DMC_READ_CHAR:
	if (copy_from_user((void *)&iobuf, (const void *)arg, sizeof(iobuf)))
	    return -EFAULT;
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 get byte at %x+%#x\n",
	    					dev->base,iobuf.offset);
	#endif
	if (iobuf.offset < 0 || iobuf.offset > DMC_MAX_OFFSET)
	    return -EINVAL;
	addr = dev->base + iobuf.offset;
	iobuf.c = inb(addr);
	if (copy_to_user((void *)arg, (void *)&iobuf, sizeof(iobuf)))
	    return -EFAULT;
	break;
    case DMC_WRITE_CHAR:
	if (copy_from_user((void *)&iobuf, (const void *)arg, sizeof(iobuf)))
	    return -EFAULT;
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 put byte %d (%#x, %#o) at %x + %#x\n",
		iobuf.c, iobuf.c, iobuf.c, dev->base, iobuf.offset);
	#endif
	if (iobuf.offset < 0 || iobuf.offset > DMC_MAX_OFFSET)
	    return -EINVAL;
	outb_p(iobuf.c, dev->base + iobuf.offset);
	if (copy_to_user((void *)arg, (void *)&iobuf, sizeof(iobuf)))
	    return -EFAULT;
	break;
    case DMC_NEED_HALFBUFFER:
	/* If arg !0, then poll() doesn't call device writable unless
	 * the buffer is at least half-empty.
	 */
	handle->need_halfbuffer = arg;
	break;
    case DMC_GET_STATS:
	#ifdef DEBUG_IOCTL
	    printk(KERN_INFO "DMC1800 get stats\n");
	#endif
	if (copy_to_user((void *)arg, (void *)&dev->stats,
						sizeof(struct dmc_stats)))
	    return -EFAULT;
	break;
    case DMC_IRQ_AVAIL:
	n = dev->interrupt[0];
	if (copy_to_user((void *)arg, (void *)&n, sizeof(int)))
	    return -EFAULT;
	break;
    case DMC_IRQ_CAUGHT:
	if (copy_to_user((void *)arg, (void *)&dev->stats.irq_caught,
						sizeof(dev->stats.irq_caught)))
	    return -EFAULT;
	break;
    case DMC_GET_INTERRUPTS:
	n = dev->interrupt[0];
	if (copy_to_user((void *)arg, (void *)dev->interrupt, n+1))
	    return -EFAULT;
	dev->interrupt[0] = 0;
	break;
    case DMC_CLEAR_STATS:
	dmc_clear_stats(dev);
	break;
    case DMC_JIFFIES:
	n = jiffies;
	if (copy_to_user((void *)arg, (void *)&n, sizeof(int)))
	    return -EFAULT;
	break;
    default:
	printk(KERN_INFO "DMC1800 ioctl: illegal command cmd=%d", cmd);
	return -EINVAL;
    }

    return 0;
}

static unsigned int
dmc_poll (
    struct file *file,
    poll_table *wait)
{
    unsigned int mask = 0;
    unsigned char cntl_byte;
    DMC1800_handle *handle = file->private_data;
    DMC1800 *dev = handle->Dmc1800;

    poll_wait(file, &dev->pollq,  wait);

    cntl_byte = inb(CNTL_ADDR(dev));
    if (handle->binary) {
	/*
	 * Binary interface
	 */
	if ((cntl_byte & DMC1800_2ND_EMPTY)) {
	    /* Secondary FIFO empty */
	    ;
	} else {
	    /* Secondary FIFO not empty */
	    mask |= (POLLIN | POLLRDNORM);	/* readable */
	}

    } else {
	/*
	 * ASCII interface
	 */

	/* are there bytes to be read? */
	if ((cntl_byte & DMC1800_READ_EMPTY) == 0)
	    mask |= (POLLIN | POLLRDNORM);	/* readable */

	/* Writable? */
	if ((cntl_byte & DMC1800_WRITE_FULL) == 0) {
	    /* Can write at least one byte */
	    if (!handle->need_halfbuffer) {
		mask |= (POLLOUT | POLLWRNORM);	/* 1 byte is enough: writable */
	    } else if ((cntl_byte & DMC1800_HALF_FULL) == 0) {
		mask |= (POLLOUT | POLLWRNORM);	/* is half-empty: writable */
	    }
	}
    }

    /* Interrupts received but not handled by user-space code? If so,
     * we'll answer to POLLPRI.
     */
    if (dev->interrupt[0])
	mask |= POLLPRI;

    return mask;
}


static int
dmc_read(
    struct file *file,
    char *buf,
    size_t nreq,
    loff_t *offset)
{
    DMC1800_handle *handle = file->private_data;
    if (handle->binary) {
	return dmc_read_binary(file, buf, nreq, offset);
    } else {
	return dmc_read_ascii(file, buf, nreq, offset);
    }
    /* NOTREACHED */
}

static int
dmc_read_ascii(
    struct file *file,
    char *buf,
    size_t nreq,
    loff_t *offset)
{
    char local;
    int nread;
    unsigned long jiffies_0;
    DMC1800 *dev = ((DMC1800_handle *) file->private_data)->Dmc1800;

    #ifdef DEBUG_READ
	printk(KERN_INFO "DMC1800 read  request nreq=%d\n",nreq);
    #endif

    dev->stats.aread++;
    jiffies_0 = jiffies;
    while (inb(CNTL_ADDR(dev)) & DMC1800_READ_EMPTY) {
	/*
	 * Read FIFO is empty.
	 */
	dev->stats.readempty++;

	if (file->f_flags & O_NONBLOCK) {
	    dev->stats.rlastsleep = 0;
	    return -EAGAIN;
	}
	/* Wait for the bh_iochecker task to awaken us. */
	dev->stats.aread_sleeping++;
	interruptible_sleep_on(&dev->a_inq);
	dev->stats.aread_sleeping--;
	if (signal_pending(current))
	    return -ERESTARTSYS;	/* tell the fs layer to handle it */
    }
    dev->stats.rlastsleep = jiffies - jiffies_0;

    /*
     * OK, have some data.
     * Read as much as available, up to requested number.
     */
    nread = 0;
    do {
	local=inb(FIFO_ADDR(dev));
	#ifdef DEBUG_READ
	{
	    printk(KERN_INFO "DMC1800 read: CNTL_ADDR=%#x  c=%#o\n",
	    					CNTL_ADDR(dev), local);
	}
	#endif
	nread++;
	if (copy_to_user((void *) buf++, (void *) &local, 1)) 
	    return -EFAULT;
    } while (nread < nreq && !(inb(CNTL_ADDR(dev)) & DMC1800_READ_EMPTY));
    #ifdef DEBUG_READ
	printk(KERN_INFO "DMC1800 read  nread=%d\n", nread);
    #endif

    if (nread == nreq)
	dev->stats.areadsatisfied++;
    return nread;
}

static int
dmc_read_binary(
    struct file *file,
    char *buf,
    size_t nreq,
    loff_t *offset)
{
    char *p, local[512];
    int nread;
    unsigned long jiffies_0;
    DMC1800 *dev = ((DMC1800_handle *) file->private_data)->Dmc1800;

    #ifdef DEBUG_READ
	printk(KERN_INFO "DMC1800 read BINARY nreq=%d\n", nreq);
    #endif

    dev->stats.bread++;
    jiffies_0 = jiffies;
    while ((inb(CNTL_ADDR(dev)) & DMC1800_2ND_EMPTY)) {
	/*
	 * Secondary FIFO is empty.
	 */
	dev->stats.readempty++;

	if (file->f_flags & O_NONBLOCK) {
	    dev->stats.rlastsleep = 0;
	    return -EAGAIN;
	}
	/* Wait for the bh_iochecker task to awaken us. */
	dev->stats.bread_sleeping++;
	interruptible_sleep_on(&dev->b_inq);
	dev->stats.bread_sleeping--;
	if (signal_pending(current))
	    return -ERESTARTSYS;
    }
    dev->stats.rlastsleep = jiffies - jiffies_0;
    
    /* Tell DMC1800 to stop updating the data record -- set freeze bit.
     * (Remember to keep IRQ's enabled.)
     */
    local[0] = DMC1800_IRQ_ENABLE | DMC1800_FREEZE_FIFO;
    outb_p(local[0], CNTL_ADDR(dev));

    /* Wait until controller has stopped updating the record */
    while ((inb(CNTL_ADDR(dev)) & DMC1800_BUSY)) {
	dev->stats.bread_sleeping++;
	interruptible_sleep_on(&dev->b_inq);
	dev->stats.bread_sleeping--;
	if (signal_pending(current))
	    return -ERESTARTSYS;
    }

    /* Read available data, up to number of requested bytes */
    if (nreq > sizeof(local))
	nreq = sizeof(local);		/* but not more than our local buffer */
    for (nread=0, p=local; nread < nreq &&
    		((inb(CNTL_ADDR(dev)) & DMC1800_2ND_EMPTY)) == 0; nread++) {
	*p++ = inb(SCND_FIFO_ADDR(dev));
    }
    if (copy_to_user((void *) buf, (void *) local, nread)) 
	return -EFAULT;

    /* Unset freeze bit -- allow DMC1800 to start updating the data record.
     * (Remember to keep IRQ's enabled.)
     */
    local[0] = DMC1800_IRQ_ENABLE | DMC1800_UNFREEZE_FIFO;
    outb_p(local[0], CNTL_ADDR(dev));

    if (nread == nreq)
	dev->stats.breadsatisfied++;
    return nread;
}


static int 
dmc_write(
    struct file *file,
    const char *buf,
    size_t nreq,
    loff_t *offset)
{
    int n, nleft, status;
    char *p, *e, local[128];
    unsigned long jiffies_0;
    DMC1800 *dev = ((DMC1800_handle *) file->private_data)->Dmc1800;

    #ifdef DEBUG_WRITE
	printk(KERN_INFO "DMC1800 write nreq=%d \n", nreq);
    #endif

    dev->stats.nwrite++;

    if (nreq == 0)
	return 0;
    
    /* If buffer is full, either return early (O_NONBLOCK) or
     * sit in a blocking loop.
     */
    jiffies_0 = jiffies;
    for (status = inb(CNTL_ADDR(dev)); (status & DMC1800_WRITE_FULL);
    						status = inb(CNTL_ADDR(dev))) {
	/*
	 * Don't write: buffer full
	 */
	dev->stats.cantwrite++;
	if (file->f_flags & O_NONBLOCK) {
	    dev->stats.wlastsleep = 0;
	    return -EAGAIN;
	}

	/* Wait for the bh_iochecker task to awaken us when we can write. */
	dev->stats.write_sleeping++;
	interruptible_sleep_on(&dev->outq);
	dev->stats.write_sleeping--;
	if (signal_pending(current))
	    return -ERESTARTSYS;
    }
    dev->stats.wlastsleep = jiffies - jiffies_0;

    /*
     * OK, can write.  Write until we are done or until buffer
     * is full again.
     */
    for (nleft=nreq; nleft > 0 && !(status & DMC1800_WRITE_FULL);
				    nleft -= n, status = inb(CNTL_ADDR(dev))) {
	if ((status & DMC1800_HALF_FULL)) {
	    /*
	     * Write FIFO is at least half-full; only write one byte.
	     */
	    n = 1;
	    if (copy_from_user((void *)local, (const void *)buf++, 1))
		return -EFAULT;
	    outb_p(local[0], FIFO_ADDR(dev));
	} else {
	    /*
	     * Buffer is at least half-empty; write up to 128 bytes.
	     */
	    n = (nleft < 128) ? nleft : 128;
	    if (copy_from_user((void *)local, (const void *)buf, n))
		return -EFAULT;
	    buf += n;
	    for (p=local, e=p+n; p < e; p++) {
		outb_p(*p, FIFO_ADDR(dev));
	    }
	}
    }
    /* finally, awake any readers */
    wake_up_interruptible(&dev->a_inq);
    wake_up_interruptible(&dev->b_inq);

    if (nleft == 0)
	dev->stats.nwritesatisfied++;

    return (nreq - nleft);	/* amount written = requested - remainder */
}

static void
dmc_clear_reg(
    DMC1800 *dev,
    unsigned long arg)
{
    unsigned char c;

    /*
     * Clear the read/write FIFO's.
     */
    if (arg & DMC_CLEAR_READ) {
	c = DMC1800_RESET_READ;
	outb_p(c, IRQ_ADDR(dev));
    }
    if (arg & DMC_CLEAR_WRITE) {
	c = DMC1800_RESET_WRITE;
	outb_p(c, IRQ_ADDR(dev));
    }
    if (arg & DMC_CLEAR_IRQ) {
	/*
	 * Enable/clear the IRQ's.
	 */
	dev->interrupt[0] = 0;

	inb(IRQ_ADDR(dev));
	c = inb(CNTL_ADDR(dev));
	c |= DMC1800_IRQ_CLEAR;
	outb_p(c, CNTL_ADDR(dev));
	c = DMC1800_IRQ_ENABLE;
	outb_p(c, CNTL_ADDR(dev));

    }

    #ifdef DEBUG_OTHER
	printk(KERN_INFO "DMC1800 clear\n");
    #endif
}

static int dmc_read_proc(
    char *page,
    char **start,
    off_t offset,
    int count,
    int *eof,
    void *data)
{
    char *ptr = page;
    size_t len;
    int i, j;
    DMC1800 *dmcp = Dmc1800List;

    for (dmcp = Dmc1800List; dmcp; dmcp = dmcp->next) {
	ptr += sprintf(ptr, "DMC1800 instance=%d\n", dmcp->instance);
	ptr += sprintf(ptr, "\tbase = %#x  n_open = %d\n",
						dmcp->base, dmcp->n_open);
	ptr += sprintf(ptr, "\tinterrupt count = %d; vector = ",
							dmcp->interrupt[0]);
	for (i = 1; i <= dmcp->interrupt[0]; i++) {
	    ptr += sprintf(ptr, "%#x ", dmcp->interrupt[i]);
	}
	ptr += sprintf(ptr, "\n");
	ptr += sprintf(ptr, "\tpollcount %ld\n", dmcp->stats.pollcount);
	ptr += sprintf(ptr, "\taread %ld areadsatisfied %ld  "
					"bread %ld breadsatisfied %ld\n",
			dmcp->stats.aread, dmcp->stats.areadsatisfied,
			dmcp->stats.bread, dmcp->stats.breadsatisfied);
	ptr += sprintf(ptr, "\taread_sleeping %ld bread_sleeping %ld\n",
			dmcp->stats.aread_sleeping, dmcp->stats.bread_sleeping);
	ptr += sprintf(ptr, "\treadempty %ld rlastsleep %ld\n",
			dmcp->stats.readempty, dmcp->stats.rlastsleep);
	ptr += sprintf(ptr, "\tnwrite %ld nwritesatisfied %ld\n",
			dmcp->stats.nwrite, dmcp->stats.nwritesatisfied);
	ptr += sprintf(ptr,
			"\tcantwrite %ld write_sleeping %ld wlastsleep %ld\n",
			dmcp->stats.cantwrite, dmcp->stats.write_sleeping,
			dmcp->stats.wlastsleep);
	ptr += sprintf(ptr, "\tirq_caught %ld\n", dmcp->stats.irq_caught);
	ptr += sprintf(ptr, "\tBH hist: ");
	for (j=0; j < NJIFFIEHIST; j++) {
	    ptr += sprintf(ptr, "%ld ", dmcp->jiffiehist[j]);
	}
	ptr += sprintf(ptr, "\n");
    }

#if 1
    len = strlen(page);
    *start = 0;
    if (offset + count >= len) {
	*eof = 1;
    } else {
	*eof = 0;
    }
#else
    *eof = (len <= offset + count) ? 1 : 0;
    *start = page + offset;
    len -= offset;
    if (len > count) len = count;
    if (len < 0) len = 0;
#endif
    return len;
}
