#ifndef DMC1800_H
#define DMC1800_H

/*  
 * We need typedefs for {,u_}int{8,16,32}_t.
 *
 * Due to bad design in some of the Linux /usr/include/ files,
 * we can't blithely #include <sys/types.h> -- there are redefines
 * all over the place.
 *
 * Therefore the user of dmc1800 must explicitly include some .h file
 * that defines {,u_}int{8,16,32}_t.  This might be <stdint.h> or
 * <sys/types.h> or similar.
 */

/* ----------------------------------------- */
#include <endian.h>
#if __BYTE_ORDER == __BIG_ENDIAN
    #define SWAP_BYTES
    #define dmctohl(x)	__bswap_32 (x)
    #define dmctohs(x)	__bswap_16 (x)
    #define hltodmc(x)	__bswap_32 (x)
    #define hstodmc(x)	__bswap_16 (x)
#else
#  if __BYTE_ORDER == __LITTLE_ENDIAN
    #define dmctohl(x)	(x)
    #define dmctohs(x)	(x)
    #define hltodmc(x)	(x)
    #define hstodmc(x)	(x)
#  endif
#endif

/* ----------------------------------------- */

/* Registers: let N be the base address.  Then the register addresses are:
 *	Read FIFO		= N
 *	Write FIFO		= N
 *	Secondary FIFO		= N+12
 *	Control			= N+4
 *	IRQ/Reset		= N+8
 */
#define FIFO_ADDR(dev)		((dev)->base)	/* read/write FIFO's */
#define CNTL_ADDR(dev)		((dev)->base+4)	/* control & status register */
#define IRQ_ADDR(dev)		((dev)->base+8)	/* irq status / reset */
#define SCND_FIFO_ADDR(dev)	((dev)->base+12) /* secondary FIFO addr */

/*
 * Maximum allowed offset to DMC_READ_CHAR / DMC_WRITE_CHAR ioctl().
 */
#define DMC_MAX_OFFSET 12

/* Depth of our internal IRQ register stack -- we will store up to this
 * many values w/o loss; it's up to caller to fetch them before we
 * start dropping them on the floor.
 */
#define DMC1800_N_INTERRUPT 32

	    /* ***************************** */
	    /* Reading control register bits */
	    /* ***************************** */
#define DMC1800_WRITE_FULL	(1)       /* 1: buffer full; don't write */
#define DMC1800_HALF_FULL	(1 << 1)  /* 1: >= half full; write 1 byte ok */
#define DMC1800_READ_EMPTY	(1 << 2)  /* 1: no data to read */
#define DMC1800_BUSY		(1 << 3)  /* Busy status of secondary FIFO */
#define DMC1800_FREEZE		(1 << 4)  /* Freeze status of secondary FIFO */
#define DMC1800_IRQ_STATUS	(1 << 5)  /* IRQ status */
#define DMC1800_IRQ_ISENABLED	(1 << 6)  /* 1: IRQ's are enabled */
#define DMC1800_2ND_EMPTY	(1 << 7)  /* 1: secondary FIFO empty */

	    /* ***************************** */
	    /* Writing control register bits */
	    /* ***************************** */
#define DMC1800_UNFREEZE_FIFO	(0     )  /* Unfreeze secondary FIFO */
#define DMC1800_FREEZE_FIFO	(1 << 4)  /* Freeze secondary FIFO */
#define DMC1800_IRQ_CLEAR	(1 << 5)  /* Clear IRQ */
#define DMC1800_IRQ_ENABLE	(1 << 6)  /* IRQ enable */

	    /* ***************************** */
	    /* Writing IRQ register bits */
	    /* ***************************** */
#define DMC1800_RESET_READ	(1 << 1)  /* 1: Reset DMC-to-PC (read) FIFO */
#define DMC1800_RESET_WRITE	(1 << 2)  /* 1: Reset PC-to-DMC (write) FIFO */
#define DMC1800_RESET_CNTLR	(1 << 7)  /* 1: Reset Controller */


    /* ******************************************************* */
			/* IRQ values */
    /* ******************************************************* */

#define DMC1800_IRQ_NONE		0x00	/* no interrupt has occurred */
#define DMC1800_IRQ_WATCHDOG_TIMER	0xD9	/* watchdog timer activated */
#define DMC1800_IRQ_COMMAND_DONE	0xDA	/* Command is done */
#define DMC1800_IRQ_PROGRAM_DONE	0xDB	/* A task (thread) is done */

#define DMC1800_IS_USER_INTERRUPT(i) ((i) & 0xFF)
#define DMC1800_IRQ_USER_0		0xF0	/* user interrupt 0 */
#define DMC1800_IRQ_USER_1		0xF1	/* user interrupt 1 */
#define DMC1800_IRQ_USER_2		0xF2	/* user interrupt 2 */
#define DMC1800_IRQ_USER_3		0xF3	/* user interrupt 3 */
#define DMC1800_IRQ_USER_4		0xF4	/* user interrupt 4 */
#define DMC1800_IRQ_USER_5		0xF5	/* user interrupt 5 */
#define DMC1800_IRQ_USER_6		0xF6	/* user interrupt 6 */
#define DMC1800_IRQ_USER_7		0xF7	/* user interrupt 7 */
#define DMC1800_IRQ_USER_8		0xF8	/* user interrupt 8 */
#define DMC1800_IRQ_USER_9		0xF9	/* user interrupt 9 */
#define DMC1800_IRQ_USER_A		0xFA	/* user interrupt A */
#define DMC1800_IRQ_USER_B		0xFB	/* user interrupt B */
#define DMC1800_IRQ_USER_C		0xFC	/* user interrupt C */
#define DMC1800_IRQ_USER_D		0xFD	/* user interrupt D */
#define DMC1800_IRQ_USER_E		0xFE	/* user interrupt E */
#define DMC1800_IRQ_USER_F		0xFF	/* user interrupt F */

#define DMC1800_IS_INPUT_INTERRUPT(i) ((i) >= DMC1800_IRQ_INPUT_1 && \
					(i) <= DMC1800_IRQ_INPUT_8)
#define DMC1800_IRQ_INPUT_1		0xE1	/* input interrupt 1 */
#define DMC1800_IRQ_INPUT_2		0xE2	/* input interrupt 2 */
#define DMC1800_IRQ_INPUT_3		0xE3	/* input interrupt 3 */
#define DMC1800_IRQ_INPUT_4		0xE4	/* input interrupt 4 */
#define DMC1800_IRQ_INPUT_5		0xE5	/* input interrupt 5 */
#define DMC1800_IRQ_INPUT_6		0xE6	/* input interrupt 6 */
#define DMC1800_IRQ_INPUT_7		0xE7	/* input interrupt 7 */
#define DMC1800_IRQ_INPUT_8		0xE8	/* input interrupt 8 */

#define DMC1800_IRQ_LIMIT_SWITCH	0xC0	/* Limit switch occurred */
#define DMC1800_IRQ_EXCESS_POS_ERR	0xC8	/* Excess position error */
#define DMC1800_IRQ_ALL_MOTION_DONE	0xD8	/* All axis motion complete */
#define DMC1800_IRQ_E_AXIS_DONE		0xD7	/* E axis motion complete */
#define DMC1800_IRQ_F_AXIS_DONE		0xD6	/* F axis motion complete */
#define DMC1800_IRQ_G_AXIS_DONE		0xD5	/* G axis motion complete */
#define DMC1800_IRQ_H_AXIS_DONE		0xD4	/* H axis motion complete */
#define DMC1800_IRQ_W_AXIS_DONE		0xD3	/* W axis motion complete */
#define DMC1800_IRQ_Z_AXIS_DONE		0xD2	/* Z axis motion complete */
#define DMC1800_IRQ_Y_AXIS_DONE		0xD1	/* Y axis motion complete */
#define DMC1800_IRQ_X_AXIS_DONE		0xD0	/* X axis motion complete */


	    /* ***************************** */
	    /* ioctls for the DMC 1800. */
	    /* ***************************** */

#define DMC_CLEAR		107	/* Clear stuff; argument bits are: */
    #define DMC_CLEAR_READ	 01	/* Clear READ buffer of primary FIFO */
    #define DMC_CLEAR_WRITE	 02	/* Clear WRITE buffer of primary FIFO */
    #define DMC_CLEAR_IRQ	 04	/* Clear/enable IRQ's; also clears
					 * internal irq vector. */
#define DMC_GET_FIFO		110	/* Read primary FIFO */
#define DMC_GET_FIFO2		111	/* Read secondary FIFO */
#define DMC_GET_CNTL		112	/* Read control register */
#define DMC_GET_IRQ		113	/* Read the IRQ register */
#define DMC_RESET_CONTROLLER	114	/* Completely reset controller */
#define DMC_READ_CHAR		115	/* Read a byte from any addr */
#define DMC_WRITE_CHAR		116	/* Write a byte to any addr */
#define DMC_GET_STATS		117	/* Get the stats struct */
#define DMC_IRQ_AVAIL		118	/* Get the num of avail interrupts */
#define DMC_IRQ_CAUGHT		119	/* Get the num of caught interrupts */
#define DMC_GET_INTERRUPTS	120	/* Read the interrupt array */
#define DMC_CLEAR_STATS		121	/* Clear the stats struct */
#define DMC_NEED_HALFBUFFER	122	/* poll() says writable iff >=halfbuf */
#define DMC_JIFFIES		123	/* Return the current jiffies */

/* The dmc_iobuf is used in the DMC_READ_CHAR and DMC_WRITE_CHAR ioctls.
 * For READ, the character at (base_address + .offset) is put into .c.
 * For WRITE, the character in .c is written to (base_address + .offset).
 */
struct dmc_iobuf {		/* used with read/write ioctl's. */
    int offset;				/*	relative to base addr */
    unsigned int c;			/*	read into this or write it */
};

/* The following statistics are kept by the device driver.
 * They can be fetched with a DMC_GET_STATS call.
 * This can be useful in analyzing I/O to the DMC 1800, thus letting
 * user-level code be optimized.
 */

struct dmc_stats {
    unsigned long pollcount;		/* number of times the poller task
    					 * for this device has been invoked.
					 */
    unsigned long aread;		/* number of ascii read calls */
    unsigned long bread;		/* number of binary read calls */
    unsigned long areadsatisfied;	/* number of ascii read calls with all
    					 * requested bytes available.
					 */
    unsigned long breadsatisfied;	/* number of binary read calls with all
    					 * requested bytes available.
					 */
    unsigned long aread_sleeping;	/* number of sleepers on a_inq */
    unsigned long bread_sleeping;	/* number of sleepers on b_inq */
    unsigned long readempty;		/* number of reads w/ in-FIFO empty */
    unsigned long rlastsleep;		/* duration (jiffies) of last block in
    					 * blocking read call.
					 */
    unsigned long nwrite;		/* number of write calls */
    unsigned long nwritesatisfied;	/* number of write calls with all
    					 * requested bytes written.
					 */
    unsigned long cantwrite;		/* number of write tries w/ full FIFO */
    unsigned long wlastsleep;		/* duration (jiffies) of last block in
    					 * blocking write call.
					 */
    unsigned long write_sleeping;	/* number of sleepers on outq */
    unsigned long irq_caught;		/* number of irq's we've caught */
};

/* ----------------------------------------- */

typedef unsigned char UINT8;
typedef unsigned short UINT16;
typedef unsigned long UINT32;

typedef signed char INT8;
typedef short INT16;
typedef long INT32;

#define DMC1800_GENERALIO	10	/* No. of general input/output bytes */
struct SecondaryFIFOheader {
    UINT16	sample_number;
    UINT8	input[DMC1800_GENERALIO];	/* general inputs. */
    UINT8	output[DMC1800_GENERALIO];	/* general outputs. */
    				/* Note: not all the input (output)
				 * bits may be available be available as
				 * inputs (outputs); "CO" selects which
				 * bits are inputs and which are outputs.
				 */
    UINT8	error_code;
    UINT8	status;
    UINT16	S_seg_count;
    UINT16	S_coordinated_move_status;
    INT32	S_dist_travelled;
    UINT16	T_seg_count;
    UINT16	T_coordinated_move_status;
    INT32	T_dist_travelled;
};

struct SecondaryFIFOaxis {
    UINT16	status;
    UINT8	switches;
    UINT8	stopcode;
    INT32	reference_position;
    INT32	motor_position;
    INT32	position_error;
    INT32	auxiliary_position;
    INT32	velocity;	/* 64 x TV value */
    INT16	torque;		/* range +-32767 <--> +-9.998 */
    INT16	analog_input;	/* Galil does this weird thing: the i'th (1..8)
    				 * analog input is available on the (i-1)th
				 * axis struct.  This is a weird implementation,
				 * because there are 8 analog inputs, even if
				 * you don't have 8 axes, but you can only
				 * see as many of them (in the secondary fifo)
				 * as you have axes!
				 */

};

/*
 * The general-purpose version.
 */
struct SecondaryFIFO {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis *a;
};
typedef struct SecondaryFIFO SecondaryFIFO;

/*
 * Versions with specific numbers of axes.
 */
struct SecondaryFIFO_1 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[1];
};
typedef struct SecondaryFIFO_1 SecondaryFIFO_1;

struct SecondaryFIFO_2 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[2];
};
typedef struct SecondaryFIFO_2 SecondaryFIFO_2;

struct SecondaryFIFO_3 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[3];
};
typedef struct SecondaryFIFO_3 SecondaryFIFO_3;

struct SecondaryFIFO_4 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[4];
};
typedef struct SecondaryFIFO_4 SecondaryFIFO_4;

struct SecondaryFIFO_5 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[5];
};
typedef struct SecondaryFIFO_5 SecondaryFIFO_5;

struct SecondaryFIFO_6 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[6];
};
typedef struct SecondaryFIFO_6 SecondaryFIFO_6;

struct SecondaryFIFO_7 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[7];
};
typedef struct SecondaryFIFO_7 SecondaryFIFO_7;

struct SecondaryFIFO_8 {
    struct SecondaryFIFOheader h;
    struct SecondaryFIFOaxis a[8];
};
typedef struct SecondaryFIFO_8 SecondaryFIFO_8;


/* ***************************************************** */
/* Bits in general_status */
/* ***************************************************** */
#define PROGRAM_RUNNING	7
#define WAITING_FOR_IN	2
#define TRACE_ON	1
#define ECHO_ON		0
/* Return the value of a header status bit */
#define HEADER_STATUS(fifo,bit) (((fifo).h.status & (1<<bit)) != 0)

/* ***************************************************** */
/* Stopcodes (axis[X].stopcode) */
/* ***************************************************** */
#define SC_RUNNING		  0	/* Motors are running, indep mode */
#define SC_INDEP_MOVE_DONE	  1	/* Motors stopped at cmd'd indep posn */
#define SC_FWD_LIMIT		  2	/* decel or stop due to FWD limit sw */
#define SC_REV_LIMIT		  3	/* decel or stop due to REV limit sw */
#define SC_ST_COMMAND		  4	/* decel or stop due to ST command */
#define SC_ABORT_INP		  6	/* stop due to Abort Input */
#define SC_AB_COMMAND		  7	/* stop due to AB command */
#define SC_OFF_ON_ERROR		  8	/* decel or stop due to OE 1 */
#define SC_FE_COMMAND_DONE	  9	/* stop after FE done */
#define SC_HM_COMMAND_DONE	 10	/* stop after HM done */
#define SC_SELECTIVE_ABORT_INP	 11	/* stop due to selective abort input */
#define SC_CONTOUR_RUNNING	 50	/* Contour running */
#define SC_CONTOUR_STOP		 51	/* Contour Stop */
#define SC_MC_TIMEOUT		 99	/* MC timeout (see MC, TW commands) */
#define SC_RUNNING_VECTOR_SEQ	100	/* Motors running, vector sequence */
#define SC_FINISHED_VECTOR_SEQ	101	/* Motors stopped at commanded vector */

/* ***************************************************** */
/* Bits in axis[X].status */
/* ***************************************************** */
#define MOVE_IN_PROGRESS	15
#define MODE_PA_OR_PR		14
#define MODE_PA_NOT_PR		13
#define FE_IN_PROGRESS		12
#define HM_IN_PROGRESS		11
#define HM_PHASE1		10
#define HM_PHASE2_OR_FI_ISSUED	 9
#define MODE_COORD		 8
#define NEGATIVE_DIRECTION	 7
#define MODE_CONTOUR		 6
#define MOTION_IS_SLEWING	 5
#define MOTION_STOP_ST_OR_LIM	 4	/* motion stop due to ST or limit */
#define MOTION_FINAL_DECEL	 3
#define LATCH_IS_ARMED		 2
#define OFF_ON_ERROR		 1
#define MOTOR_OFF		 0

/*
 * Here are the active status bits when doing a jog (not shown: OFF_ON_ERROR,
 * which simply follows the setting of the OE command):
 *
 *"JG 40000"
 *    MOTION_IS_SLEWING; MOTION_STOP_ST_OR_LIM; MOTION_FINAL_DECEL
 * The MOTION_IS_SLEWING is the only significant bit.  The other
 * MOTION_xxx bits are simply left over from the last setting of
 * the bits.
 *
 *"BGX"
 *Start:
 * MOVE_IN_PROGRESS
 *Note that "MOTION_IS_SLEWING" is *not* active during the acceleration.
 *When full speed is reached:
 * MOVE_IN_PROGRESS; MOTION_IS_SLEWING
 *When ST is typed:
 * MOVE_IN_PROGRESS; MOTION_IS_SLEWING; MOTION_STOP_ST_OR_LIM; MOTION_FINAL_DECEL
 * MOVE_IN_PROGRESS; MOTION_STOP_ST_OR_LIM; MOTION_FINAL_DECEL
 *Note that "MOTION_IS_SLEWING" is *not* active during the deceleration.
 *When deceleration is complete:
 * MOTION_IS_SLEWING; MOTION_STOP_ST_OR_LIM; MOTION_FINAL_DECEL
 *
 *---------------------------------------------------------------------
 *
 *PR 1000000
 * MODE_PA_OR_PR; MOTION_IS_SLEWING; MOTION_STOP_ST_OR_LIM; MOTION_FINAL_DECEL
 *
 *BGX
 *Start:
 * MOVE_IN_PROGRESS; MODE_PA_OR_PR
 *when full speed is reached:
 * MOVE_IN_PROGRESS; MODE_PA_OR_PR; MOTION_IS_SLEWING
 *when time to decelerate:
 * MOVE_IN_PROGRESS; MODE_PA_OR_PR; MOTION_IS_SLEWING; MOTION_FINAL_DECEL
 * MOVE_IN_PROGRESS; MODE_PA_OR_PR; MOTION_FINAL_DECEL
 *When deceleration is complete:
 * MODE_PA_OR_PR; MOTION_IS_SLEWING; MOTION_FINAL_DECEL
 *
 *If ST is issued during the move:
 * The decelerate and motion-complete bits are the same as above, plus
 * the MOTION_STOP_ST_OR_LIM is set.
 */



/* Return the value of an axis status bit */
#define AXIS_STATUS(fifo,ax,bit) (((fifo).a[ax].status >> (bit)) & 1)

/* AND one or more bits with the axis status */
#define AXIS_STATUS_BITS(fifo,ax,bitmask) ((fifo).a[ax].status && (bitmask))

/* Return !0 if a particular axis status bit has changed */
#define AXIS_STA_BIT_CHANGED(fifo1,fifo2,ax,bit) \
		(AXIS_STATUS(fifo1,ax,bit) != AXIS_STATUS(fifo2,ax,bit))

/* */

/* ***************************************************** */
/* Bits in axis[X].switches */
/* ***************************************************** */
#define LATCH_OCCURRED		7
#define STATE_OF_LATCH_INPUT	6
#define STATE_OF_FWD_LIMIT	3
#define STATE_OF_REV_LIMIT	2
#define STATE_OF_HOME_INPUT	1
#define SM_JUMPER_INSTALLED	0

/* Return the value of an axis switch bit */
#define AXIS_SWITCH(fifo,ax,bit) (((fifo).a[ax].switches >> (bit)) & 1)

/* AND one or more bits with the axis switch */
#define AXIS_SW_BITS(fifo,ax,bitmask) ((fifo).a[ax].switches && (bitmask))

/* Return !0 if a particular axis switch bit has changed */
#define AXIS_SW_BIT_CHANGED(fifo1,fifo2,ax,bit) \
		(AXIS_SWITCH(fifo1,ax,bit) != AXIS_SWITCH(fifo2,ax,bit))
/* */

/* ***************************************************** */
/* Bits in coordinated motion status (S or T) */
/* ***************************************************** */
#define COORD_MOVE_IN_PROGRESS	15
#define COORD_MOTION_SLEWING	 5
#define COORD_STOP_ST_OR_LIM	 4
#define COORD_MOTION_FINAL_DECEL 3
/* */

/*
 * Convert Secondary FIFO velocity to counts/sec.
 */
#define FIFO_TO_TRUE_VEL(v) ((v) / 64)

/*
 * Convert an analog input value to float volts.
 */
#define ANALOG_TO_FLOAT(x) ((x) * 0.000305124057741)
/*
 * Convert a float volts to equivalent analog input integer.
 */
#define FLOAT_TO_ANALOG(x) ((x) / 0.000305124057741)

/*
 * Convert Secondary FIFO torque to true torque, and back again.
 */
#define FIFO_TO_TRUE_TORQUE(t) ((t) * 0.000305124057741)
#define TRUE_TORQUE_TO_FIFO(t) ((t) / 0.000305124057741)

/* ***************************************************** */
/* General input and output bits. */
/* ***************************************************** */

/* In these macros, the "dir" (direction) must be literally `input' or
 * `output'.  Note that it is used with the token-pasting operator
 * to create the correct (input or output) struct field.
 */

/* Convert "global" bit numbers (1..80) into block, block-bit numbers */
/* Example: bit 9 is I/O block 1, bit 0; bit 10 is I/O block 1, bit 1. */
#define IO_BIT_BLOCK(bit)	(((bit)-1) / 8)
#define IO_BIT_NUM(bit)	(((bit)-1) % 8)

/* Utility macros: these evaluate to true (false) if their argument bit
 * numbers, taken from the range 1..80, do (don't) fall into the same
 * input or output block.
 */
#define IO_BIT_SAMEBLOCK_2(bit1, bit2)					\
		(IO_BIT_BLOCK(bit1) == IO_BIT_BLOCK(bit2))

#define IO_BIT_SAMEBLOCK_3(bit1, bit2, bit3)				\
		((IO_BIT_BLOCK(bit1) == IO_BIT_BLOCK(bit2)) &&		\
		(IO_BIT_BLOCK(bit1) == IO_BIT_BLOCK(bit3)))

#define IO_BIT_SAMEBLOCK_4(bit1, bit2, bit3, bit4)			\
		((IO_BIT_BLOCK(bit1) == IO_BIT_BLOCK(bit2)) &&		\
		(IO_BIT_BLOCK(bit1) == IO_BIT_BLOCK(bit3)) &&		\
		(IO_BIT_BLOCK(bit1) == IO_BIT_BLOCK(bit4)))

/* Utility macros: *ASSUMING* that the argument bit numbers all fall in
 * the same byte (ie 1..8, 9..16, etc), we construct bitmasks for selecting
 * 2, 3, or 4 bits from an byte-sized bit block.
 *
 * The *CALLER* must ensure that the argument bit numbers actually all do
 * fall into one byte, e.g. by checking that
 * IO_BIT_BLOCK(bit1) == IO_BIT_BLOCK(bit2) == ....
 */
#define IO_BITMASK_2(bit1, bit2) ((1 << IO_BIT_NUM((bit1))) |		\
					(1 << IO_BIT_NUM((bit2))))

#define IO_BITMASK_3(b1, b2, b3) ((1 << IO_BIT_NUM((b1))) |		\
					(1 << IO_BIT_NUM((b2))) |	\
					(1 << IO_BIT_NUM((b3))))

#define IO_BITMASK_4(b1, b2, b3, b4) ((1 << IO_BIT_NUM((b1))) |		\
					(1 << IO_BIT_NUM((b2))) |	\
					(1 << IO_BIT_NUM((b3))) |	\
					(1 << IO_BIT_NUM((b4))))

#define IO_BIT_VALUE(direction, fifo, bit) \
	(((fifo).h.direction[IO_BIT_BLOCK(bit)] >> IO_BIT_NUM(bit)) & 1)

#define IO_BIT_ISSET(dir, fifo, bit) IO_BIT_VALUE(dir, (fifo), (bit))
#define IO_BIT_ISCLR(dir, fifo, bit) (!IO_BIT_VALUE(dir, (fifo), (bit)))

/* All bits are active-low */
#define IO_IS_ACTIVE(dir, fifo, bit) IO_BIT_ISCLR(dir, (fifo), (bit))

#define IO_BIT_EQUAL(dir, fifo1, bit1, fifo2, bit2) \
    (IO_BIT_VALUE(dir, (fifo1), (bit1)) == IO_BIT_VALUE(dir, (fifo2), (bit2)))

#define IO_BIT_CHANGED(dir, fifo1, fifo2, bit) \
	(!IO_BIT_EQUAL(dir, (fifo1), (bit), (fifo2), (bit)))

/* ***************************************************** */
/* Specific input and output macros. */
/* ***************************************************** */

/* These are simply the general-purpose macros, with the specific
 * input or output substitution.
 */

#define INP_BIT_VALUE(fifo, bit) IO_BIT_VALUE(input, fifo, bit)
#define INP_BIT_ISSET(fifo, bit) IO_BIT_ISSET(input, fifo, bit) 
#define INP_BIT_ISCLR(fifo, bit) IO_BIT_ISCLR(input, fifo, bit)
#define INP_IS_ACTIVE(fifo, bit) IO_IS_ACTIVE(input, fifo, bit)
#define INP_BIT_EQUAL(fifo1, bit1, fifo2, bit2) \
				IO_BIT_EQUAL(input, fifo1, bit1, fifo2, bit2)
#define INP_BIT_CHANGED(fifo1, fifo2, bit) \
				IO_BIT_CHANGED(input, fifo1, fifo2, bit)

#define OUT_BIT_VALUE(fifo, bit) IO_BIT_VALUE(output, fifo, bit)
#define OUT_BIT_ISSET(fifo, bit) IO_BIT_ISSET(output, fifo, bit) 
#define OUT_BIT_ISCLR(fifo, bit) IO_BIT_ISCLR(output, fifo, bit)
#define OUT_IS_ACTIVE(fifo, bit) IO_IS_ACTIVE(output, fifo, bit)
#define OUT_BIT_EQUAL(fifo1, bit1, fifo2, bit2) \
				IO_BIT_EQUAL(output, fifo1, bit1, fifo2, bit2)
#define OUT_BIT_CHANGED(fifo1, fifo2, bit) \
				IO_BIT_CHANGED(output, fifo1, fifo2, bit)

/* ***************************************************** */
/* Binary command structure */
/* ***************************************************** */
/*
 * A binary command consists of a 4-byte header, followed if
 * necessary by data values.  The data values can be 2-byte ints,
 * 4-byte ints, or 6-byte reals.
 */
struct BinaryCmd {
    unsigned char cmd;			/* The command, using one of the
    					 * below BINARY_xx codes.
					 */
    unsigned char datasize;		/* The size of each data value:
    					 * 0 if no values;
					 * 2 = 2-byte integer values;
					 * 4 = 4-byte integer values;
					 * 6 = 6-byte real (see manual for
					 *	details).
					 */
    unsigned char is_coordinated_move;	/* boolean: 0 if not a coordinated
    					 * move; 1 if it is a coord move.
					 */
    unsigned char axes;			/* bit 0..7 mean axes A..H */
    union {
	INT16 i16;			/* for holding a 2-byte int value;
					 * use hs2dmc(k) to convert a host
					 * short to the right byte order:
					 * xxx.value.i16 = hs2dmc(nn);
					 */
	INT32 i32;			/* for holding a 4-byte int value;
					 * use hl2dmc(k) to convert a host
					 * long to the right byte order:
					 * xxx.value.i32 = hl2dmc(nnn);
					 */
	unsigned char r[6];		/* for holding a 6-byte real value */
    } value;
};

/* ***************************************************** */
/* Binary codes for various commands */
/* ***************************************************** */

#define BINARY_KP	0x81	/* Propotional Constant */
#define BINARY_KI	0x82	/* Integrator */
#define BINARY_KD	0x83	/* Derivative Constant */
#define BINARY_DV	0x84	/* Dual Velocity (Dual Loop) */
#define BINARY_ER	0x88	/* Error Limit */
#define BINARY_IL	0x89	/* Integrator Limit */
#define BINARY_TL	0x8A	/* Torque Limit */
#define BINARY_MT	0x8B	/* Motor Type */
#define BINARY_CE	0x8C	/* Configure Encoder */
#define BINARY_OE	0x8D	/* Off on Error */
#define BINARY_FL	0x8E	/* Forward Software Limit */
#define BINARY_BL	0x8F	/* Reverse Software Limit */

#define BINARY_AC	0x90	/* Acceleration */
#define BINARY_DC	0x91	/* Deceleration */
#define BINARY_SP	0x92	/* Speed */
#define BINARY_IT	0x93	/* Independent Time Constant - Smoothing Function */
#define BINARY_FA	0x94	/* Acceleration Feedforward */
#define BINARY_FV	0x95	/* Velocity Feedforward */
#define BINARY_GR	0x96	/* Gear Ratio */
#define BINARY_DP	0x97	/* Define Position */
#define BINARY_DE	0x98	/* Dual (Auxiliary) Encoder Position */
#define BINARY_OF	0x99	/* Offset */

#define BINARY_BG	0xA0	/* Begin */
#define BINARY_ST	0xA1	/* Stop */
#define BINARY_AB	0xA2	/* Abort */
#define BINARY_HM	0xA3	/* Home */
#define BINARY_FE	0xA4	/* Find Edge */
#define BINARY_FI	0xA5	/* Find Index */
#define BINARY_PA	0xA6	/* Position Absolute */
#define BINARY_PR	0xA7	/* Position Relative */
#define BINARY_JG	0xA8	/* Jog */
#define BINARY_MO	0xA9	/* Motor Off */
#define BINARY_SH	0xAA	/* Servo Here */

#define BINARY_LM	0xB0	/* Linear Interpolation Mode */
#define BINARY_LI	0xB1	/* Linear Interpolation Distance */
#define BINARY_VP	0xB2	/* Vector Position */
#define BINARY_CR	0xB3	/* Circle */
#define BINARY_TN	0xB4	/* Tangent */
#define BINARY_LE	0xB5	/* Linear Interpolation End */
#define BINARY_VT	0xB6	/* Vector Time Constant - S Curve */
#define BINARY_VA	0xB7	/* Vector Acceleration */
#define BINARY_VD	0xB8	/* Vector Deceleration */
#define BINARY_VS	0xB9	/* Vector Speed */
#define BINARY_VR	0xBA	/* Vector Speed Ratio */
#define BINARY_CM	0xBD	/* Contouring Mode */
#define BINARY_CD	0xBE	/* Contour Data */
#define BINARY_DT	0xBF	/* Delta Time */

#define BINARY_ET	0xC0	/* Electronic cam table */
#define BINARY_EM	0xC1	/* Cam cycles */
#define BINARY_EP	0xC2	/* Cam table intervals and starting point */
#define BINARY_EG	0xC3	/* ECAM go (engage) */
#define BINARY_EB	0xC4	/* Enable ECAM */
#define BINARY_EQ	0xC5	/* ECAM quit (disengage) */
#define BINARY_EC	0xC6	/* ECAM table position */
#define BINARY_AM	0xC8	/* After Move */
#define BINARY_MC	0xC9	/* Motion Complete - "In Position" */
#define BINARY_TW	0xCA	/* Timeout for IN-Position (MC) */
#define BINARY_MF	0xCB	/* Forward Motion to Position */
#define BINARY_MR	0xCC	/* Reverse Motion to Position */
#define BINARY_AD	0xCD	/* After Distance */
#define BINARY_AP	0xCE	/* After Absolute Position */
#define BINARY_AR	0xCF	/* After Relative Position */

#define BINARY_AS	0xD0	/* At Speed */
#define BINARY_AI	0xD1	/* After Input */
#define BINARY_AT	0xD2	/* At Time */
#define BINARY_WT	0xD3	/* Wait */
#define BINARY_WC	0xD4	/* Wait for Contour Data */
#define BINARY_RP	0xD8	/* Reference Position */
#define BINARY_TP	0xD9	/* Tell Position */
#define BINARY_TE	0xDA	/* Tell Error */
#define BINARY_TD	0xDB	/* Tell Dual Encoder */
#define BINARY_TV	0xDC	/* Tell Velocity */
#define BINARY_RL	0xDD	/* Report Latched Position */
#define BINARY_TT	0xDE	/* Tell Torque */
#define BINARY_TS	0xDF	/* Tell Switches */

#define BINARY_TI	0xE0	/* Tell Inputs */
#define BINARY_SC	0xE1	/* Stop Code */
#define BINARY_TM	0xE5	/* Time */
#define BINARY_CN	0xE6	/* Configure */
#define BINARY_LZ	0xE7	/* Leading Zeros */
#define BINARY_OP	0xE8	/* Output Port */
#define BINARY_OB	0xE9	/* Output Bit */
#define BINARY_SB	0xEA	/* Set Bit */
#define BINARY_CB	0xEB	/* Clear Bit */
#define BINARY_II	0xEC	/* Input Interrupt */
#define BINARY_EI	0xED	/* Enable Interrupts */
#define BINARY_AL	0xEE	/* Arm Latch */

#endif /* DMC1800_H */
